from argparse import ArgumentParser
import os
import glob

from dodo_commands import Dodo


def _args():
    parser = ArgumentParser()
    parser.add_argument('where')
    parser.add_argument('what')
    parser.add_argument('--pattern', default='*')
    parser.add_argument('--replace')
    args = Dodo.parse_args(parser)
    return args


def _replace(where, what, replace_with):
    for filepath in glob.iglob(
            os.path.join(where, '**/' + args.pattern), recursive=True):
        with open(filepath) as file:
            s = file.read()
        s2 = s.replace(what, replace_with)
        if s != s2:
            with open(filepath, "w") as file:
                file.write(s2)


args = _args()
if Dodo.is_main(__name__, safe=not args.replace):
    if args.replace:
        _replace(args.where, args.what, args.replace)
    else:
        Dodo.run(['grep', '-rnw', args.where, '-e', "'{}'".format(args.what)])



from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config import build_config, extend_command_path
from dodo_commands.framework.config_layers import get_conflicts_in_layer_paths
from dodo_commands.framework.container.facets import (Config, Layers, i_,
                                                      map_datas, o_)
from dodo_commands.framework.transform_prefixes_in_aliases import \
    transform_prefixes_in_aliases


# CONFIG
def action_check_conflicts_in_selected_layer_paths(ctr):
    def transform(selected_layer_by_path):
        conflicts = get_conflicts_in_layer_paths(selected_layer_by_path.keys())
        for path, conflicting_path in conflicts:
            raise CommandError("Conflicting layers: %s and %s" %
                               (path, conflicting_path))
        return tuple()

    return map_datas(i_(Layers, 'selected_layer_by_path'),
                     transform=transform)(ctr)


# CONFIG
def action_build_from_selected_layers(ctr):
    def transform(
        #
        selected_layer_by_path,
        layer_props_by_layer_name):
        def get_selected_layers():
            return selected_layer_by_path.values()

        def _build_config(layers):
            return build_config(layers)

        x = get_selected_layers()
        # [layer]
        config = build_config(x)
        config = extend_command_path(config)
        transform_prefixes_in_aliases(config, layer_props_by_layer_name)

        return (config, )

    return map_datas(i_(Layers, 'selected_layer_by_path'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     o_(Config, 'config'),
                     transform=transform)(ctr)



import os
import sys
from argparse import ArgumentParser
from configparser import NoOptionError

from dodo_commands import CommandError, Dodo
from dodo_commands.framework.global_config import load_global_config_parser


def _args():
    parser = ArgumentParser(description='Print the command line to cd to'
                            ' a folder inside the project folder.')
    parser.add_argument('--browse',
                        action='store_true',
                        help='Open file browser')

    parser.add_argument(
        'to',
        nargs='?',
        help=('cd to /ROOT/<to>_dir. For example: dodo cd src ' +
              'cds to the value of /ROOT/src_dir.'))
    args = Dodo.parse_args(parser)
    args.path = Dodo.get_config("/ROOT/%s_dir" % (args.to or 'project'))
    return args


if Dodo.is_main(__name__):
    args = _args()

    if args.browse:
        try:
            browser = load_global_config_parser().get("settings",
                                                      "file_browser")
        except NoOptionError:
            raise CommandError(
                "The file_browser option was not found in the global configuration."
            )
        Dodo.run([browser, args.path])

    sys.stdout.write("cd %s\n" % os.path.expanduser(args.path))



import dodo_commands.framework.container.actions.layers  # noqa
import dodo_commands.framework.container.actions.command_line  # noqa
import dodo_commands.framework.container.actions.config  # noqa
import dodo_commands.framework.container.actions.commands  # noqa



import os

import dodo_commands
from dodo_commands import CommandError
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.util import symlink


def site_packages_dir(python_env_dir):
    python = plumbum.local[os.path.join(python_env_dir, "bin", "python")]

    result = python(
        "-c", "from distutils.sysconfig import get_python_lib; " +
        "print(get_python_lib())")
    while result[-1] in ['\n', '\r']:
        result = result[:-1]
    return result


def create_python_env(python, python_env_dir, env):
    try:
        plumbum.local["virtualenv"]('--version')
    except:  # noqa
        raise CommandError(
            "Could not find virtualenv, please install it first.")

    plumbum.local["virtualenv"]("-p", python, python_env_dir, "--prompt", env)

    symlink(os.path.realpath(os.path.dirname(dodo_commands.__file__)),
            os.path.join(site_packages_dir(python_env_dir), "dodo_commands"))



"""Pauses the execution."""
import time

from dodo_commands.framework.decorator_utils import uses_decorator


class Decorator:
    def is_used(self, config, command_name, decorator_name):
        return uses_decorator(config, command_name, decorator_name)

    def add_arguments(self, parser):  # override
        parser.add_argument(
            '--pause-ms',
            type=int,
            help="Pause in milliseconds before continuing")

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # override
        if getattr(command_line_args, 'pause_ms', 0):
            time.sleep(command_line_args.pause_ms / 1000)
        return args_tree_root_node, cwd



import os


def _env():
    return os.environ.get("DODO_COMMANDS_ENV", "default")


class Paths:
    # Cached result of finding the current project dir
    _project_dir = None
    _config_dir = None

    def __init__(self):
        Paths._project_dir = Paths._project_dir or self.project_dir(_env())
        Paths._config_dir = Paths._config_dir or self.config_dir(_env())

    def home_dir(self, expanduser=True):
        return os.path.expanduser('~') if expanduser else '~'

    def envs_dir(self):
        return os.path.join(self.global_config_dir(), "envs")

    def env_dir(self, env=None):
        return os.path.join(self.envs_dir(), env or _env())

    def global_config_dir(self, expanduser=True):
        return os.path.join(self.home_dir(expanduser), '.dodo_commands')

    def global_config_filename(self):
        return os.path.join(self.global_config_dir(), 'config')

    def default_project_dir(self, expanduser=True):
        return os.path.join(self.global_config_dir(expanduser),
                            'default_project')

    def default_commands_dir(self, expanduser=True):
        return os.path.join(self.default_project_dir(), 'commands')

    def global_commands_dir(self, expanduser=True):
        return os.path.join(self.global_config_dir(expanduser), 'commands')

    def project_dir(self, env=None):
        """Return the root dir of the current project."""
        if env is None:
            return Paths._project_dir

        return os.path.realpath(os.path.join(self.env_dir(env), "project_dir"))

    def config_dir(self, env=None):
        if env is None:
            return Paths._config_dir

        return os.path.realpath(os.path.join(self.env_dir(env), "config_dir"))

    def package_dir(self):
        import dodo_commands
        return os.path.dirname(dodo_commands.__file__)

    def extra_dir(self):
        return os.path.join(self.package_dir(), "extra")



import argparse
import os

from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.container.facets import (CommandLine, Commands,
                                                      Layers, i_, map_datas,
                                                      o_)
from dodo_commands.framework.funcy import map_with
from dodo_commands.framework.handle_arg_complete import handle_arg_complete


# COMMAND LINE
def action_get_expanded_layer_paths(ctr):
    def transform(
        #
        layer_names,
        layer_props_by_layer_name,
    ):
        def map_to_path(layer_name):
            if layer_name not in layer_props_by_layer_name:
                known_layer_names = layer_props_by_layer_name.keys()
                raise CommandError("Unknown layer: %s. Known layers: %s" %
                                   (layer_name, ", ".join(known_layer_names)))
            return layer_props_by_layer_name[layer_name].target_path

        return (map_with(map_to_path)(layer_names), )

    return map_datas(i_(CommandLine, 'layer_names'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     o_(CommandLine, 'expanded_layer_paths'),
                     transform=transform)(ctr)


# COMMAND LINE
def action_get_inferred_layer_paths(ctr):
    def transform(
        #
        raw_command_name,
        layer_name_by_inferred_command,
        layer_props_by_layer_name):
        layer_name = layer_name_by_inferred_command.get(raw_command_name, None)
        layer_props = layer_props_by_layer_name.get(layer_name)

        return ([layer_props.target_path] if layer_props else [], )

    return map_datas(i_(CommandLine, 'raw_command_name'),
                     i_(Commands, 'layer_name_by_inferred_command'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     o_(CommandLine, 'inferred_layer_paths'),
                     transform=transform)(ctr)


# COMMAND LINE
def action_expand_and_autocomplete_command_name(ctr):
    def transform(
        #
        raw_command_name,
        input_args,
        command_map,
        command_aliases,
        layer_name_by_inferred_command,
        layer_props_by_layer_name,
    ):
        completed_command_name = (
            handle_arg_complete(
                command_names=list(command_map.keys()),
                inferred_command_names=list(
                    layer_name_by_inferred_command.keys()),
                command_aliases=list(command_aliases.keys()),
                layer_props_by_layer_name=layer_props_by_layer_name)
            #
            if "_ARGCOMPLETE" in os.environ else
            #
            raw_command_name)

        alias_target = (command_aliases.get(completed_command_name)
                        or completed_command_name or 'help')

        new_args = alias_target.split(' ')

        parser = argparse.ArgumentParser()
        parser.add_argument('-L', '--layer', action='append')
        known_args, new_args = parser.parse_known_args(new_args)

        command_name = new_args[0]
        input_args = input_args[:1] + new_args + input_args[2:]
        return (input_args, command_name, known_args.layer or [])

    return map_datas(i_(CommandLine, 'raw_command_name'),
                     i_(CommandLine, 'input_args'),
                     i_(Commands, 'command_map'),
                     i_(Commands, 'aliases', alt_name='command_aliases'),
                     i_(Commands, 'layer_name_by_inferred_command'),
                     i_(Layers, 'layer_props_by_layer_name'),
                     o_(CommandLine, 'input_args'),
                     o_(CommandLine, 'command_name'),
                     o_(CommandLine, 'more_given_layer_paths'),
                     transform=transform)(ctr)



import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum, six
from dodo_docker_commands.decorators.docker import Decorator as DockerDecorator

docker = plumbum.cmd.docker
raw_input = six.moves.input


def _args():
    parser = ArgumentParser()
    parser.add_argument('--user')
    parser.add_argument('--find')
    parser.add_argument('name', nargs='?')
    parser.add_argument('--cmd')
    args = Dodo.parse_args(parser)

    return args


def _containers():
    return [x for x in docker("ps", "--format", "{{.Names}}").split('\n') if x]


if Dodo.is_main(__name__):
    args = _args()

    if args.find:
        args.name = None
        for container in _containers():
            if args.find in container:
                args.name = container
                break
        if not args.name:
            raise CommandError("Container not found: %s" % args.find)
    elif not args.name:
        containers = _containers()
        print("0 - exit")
        for idx, container in enumerate(containers):
            print("%d - %s" % (idx + 1, container))

        print("\nSelect a container: ")
        choice = int(raw_input()) - 1

        if choice == -1:
            sys.exit(0)

        args.name = containers[choice]

    if not args.cmd:
        default_shell = Dodo.get_config("/DOCKER/default_shell", "sh")
        docker_options = DockerDecorator.merged_options(
            Dodo.get_config, "docker-exec")
        args.cmd = docker_options.get("shell", default_shell)

    Dodo.run([
        'docker',
        'exec',
        '-i',
        '-t',
    ] + (['--user', args.user] if args.user else []) + [
        args.name,
        args.cmd,
    ], )



import glob
import os

from dodo_commands.dependencies import (yaml_round_trip_dump,
                                        yaml_round_trip_load)
from dodo_commands.framework.paths import Paths


class ConfigIO:
    """Read and writes dodo configuration (yaml) files."""
    def __init__(self, config_base_dir=None):
        """Arg config_base_dir is where config files are searched.

        Arg config_base_dir defaults to Paths().config_dir().
        """
        self._cache = {}
        self.config_base_dir = (Paths().config_dir() if config_base_dir is None
                                else config_base_dir)

    def _path_to(self, post_fix_paths):
        """Return path composed of config_base_dir and post_fix_paths list."""
        return os.path.join(self.config_base_dir, *post_fix_paths)

    def glob(self, patterns):
        """Returns list of filenames"""
        def add_prefix(filename):
            return (filename
                    if filename.startswith('/') else self._path_to([filename]))

        if not self.config_base_dir:
            return []

        patterns = [
            add_prefix(os.path.expanduser(pattern)) for pattern in patterns
        ]

        result = []
        for pattern in patterns:
            result.extend([x for x in glob.glob(pattern)])
        return result

    def load(self, config_filename='config.yaml'):
        """Get configuration."""
        full_config_filename = self._path_to([config_filename])

        if full_config_filename in self._cache:
            return self._cache[full_config_filename]

        if not os.path.exists(full_config_filename):
            return None

        with open(full_config_filename) as f:
            config = yaml_round_trip_load(f.read())

        self._cache[full_config_filename] = config
        return config

    def save(self, config, config_filename='config.yaml'):
        """Write config to config_filename."""
        with open(self._path_to([config_filename]), 'w') as f:
            return f.write(yaml_round_trip_dump(config))



import os
import shlex
from argparse import ArgumentParser

from dodo_commands.dependencies import yaml_round_trip_load
from dodo_commands.framework.command_map import CommandMapItem
from dodo_commands.framework.config import Key, expand_keys
from dodo_commands.framework.singleton import Dodo
from dodo_commands.framework.util import chop


class YamlCommandMapItem(CommandMapItem):
    def __init__(self, group, filename):
        super(YamlCommandMapItem, self).__init__(group, filename, 'yaml')


class YamlCommandHandler:
    def add_commands_to_map(self, command_dirs, file_map, command_map):
        for command_dir, files in file_map.items():
            for file in files:
                command_name, ext = os.path.splitext(os.path.basename(file))
                prefix = 'dodo.'
                if ext == '.yaml' and command_name.startswith(prefix):
                    command_name = command_name[len(prefix):]
                    with open(file) as ifs:
                        data = yaml_round_trip_load(ifs.read())
                        for command_name in data.keys():
                            command_map[command_name] = YamlCommandMapItem(
                                group=os.path.basename(command_dir),
                                filename=file)

    def _expand(self, x):
        return os.path.expanduser(
            os.path.expandvars(expand_keys(Dodo.get_config(), x)))

    def execute(self, command_map_item, command_name):
        with open(command_map_item.filename) as ifs:
            data = yaml_round_trip_load(ifs.read())

        description = None
        for key, val in data[command_name].items():
            if key == '_description':
                description = val

        parser = ArgumentParser(description=description)

        for key, val in data[command_name].items():
            if key == '_args':
                for arg in val:
                    default_val = None
                    if arg.startswith('--') and '=' in arg:
                        arg, default_val = arg.split('=')
                    parser.add_argument(arg,
                                        default=None if default_val is None
                                        else self._expand(default_val))

        args = Dodo.parse_args(parser)
        Dodo.get_config()['_ARGS'] = args.__dict__

        for name, cmd in data[command_name].items():
            if name.startswith('_'):
                continue

            args = cmd['args']
            if isinstance(args, str):
                args = shlex.split(args)

            try:
                cwd = self._expand(cmd.get('cwd'))
            except:
                cwd = None

            store_xpath = cmd.get('store', '')
            res = Dodo.run([self._expand(x) for x in args],
                           cwd=cwd,
                           capture=bool(store_xpath))
            if store_xpath:
                key = Key(Dodo.get_config(), store_xpath)
                key.set(chop(res))



from argparse import REMAINDER, ArgumentParser

from dodo_commands import Dodo, remove_trailing_dashes
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.config import expand_keys


def _args():
    parser = ArgumentParser(description='Run docker compose')

    parser.add_argument('compose_args', nargs=REMAINDER)
    args = Dodo.parse_args(parser)
    args.cwd = Dodo.get_config('/DOCKER_COMPOSE/cwd')
    args.map = Dodo.get_config('/DOCKER_COMPOSE/map', {})
    return args


# Use safe=False if the script makes changes other than through Dodo.run
if Dodo.is_main(__name__, safe=True):
    args = _args()

    for src, dest in args.map.items():
        with open(src) as ifs:
            content = expand_keys(Dodo.get_config(), ifs.read())
        with open(dest, 'w') as ofs:
            ofs.write(content)

    compose_project_name = Dodo.get_config(
        '/DOCKER_COMPOSE/compose_project_name',
        Dodo.get_config('/ROOT/env_name'))
    with plumbum.local.env(COMPOSE_PROJECT_NAME=compose_project_name):
        Dodo.run(['docker-compose'] +
                 remove_trailing_dashes(args.compose_args),
                 cwd=args.cwd)



import argparse
import os

from dodo_commands import CommandError, Dodo, remove_trailing_dashes


def _args():
    parser = argparse.ArgumentParser()
    parser.add_argument('name',
                        choices=Dodo.get_config('/DOCKER_IMAGES').keys(),
                        help='Key to look up in /DOCKER_IMAGES')
    parser.add_argument('build_args',
                        nargs=argparse.REMAINDER,
                        help='Extra args to pass to docker build')
    args = Dodo.parse_args(parser)
    args.build_dir = Dodo.get_config(
        '/DOCKER_IMAGES/{name}/build_dir'.format(name=args.name), '.')
    args.docker_file = Dodo.get_config(
        '/DOCKER_IMAGES/{name}/docker_file'.format(name=args.name),
        'Dockerfile')
    args.extra_dirs = Dodo.get_config(
        '/DOCKER_IMAGES/{name}/extra_dirs'.format(name=args.name), [])
    args.docker_image = Dodo.get_config(
        '/DOCKER_IMAGES/{name}/image'.format(name=args.name))
    return args


def _copy_extra_dirs(local_dir, extra_dirs):
    # todo: convert extra dirs config mapping into a list of key=val
    for extra_dir in (extra_dirs or []):
        extra_dir_name, extra_dir_path = extra_dir.split('=')
        local_path = os.path.join(local_dir, extra_dir_name)
        if os.path.exists(local_path):
            raise CommandError("Cannot copy to existing path: " + local_path)
        if not os.path.exists(extra_dir_path):
            raise CommandError("Cannot copy from non-existing path: " +
                               extra_dir_path)

        Dodo.run(['cp', '-rf', extra_dir_path, local_path])
        if (os.path.abspath(local_dir).startswith(
                os.path.abspath(extra_dir_path))):
            rp = os.path.relpath(local_dir, extra_dir_path)
            dead_path = os.path.join(local_path, rp)
            if os.path.exists(dead_path):
                Dodo.run(['rm', '-rf', dead_path])


def _remove_extra_dirs(local_dir, extra_dirs):
    for extra_dir in (extra_dirs or []):
        extra_dir_name, extra_dir_path = extra_dir.split('=')
        local_path = os.path.join(local_dir, extra_dir_name)
        Dodo.run(['rm', "-rf", local_path])


if Dodo.is_main(__name__):
    args = _args()
    Dodo.safe = len(args.extra_dirs) == 0

    _copy_extra_dirs(args.build_dir, args.extra_dirs)

    try:
        Dodo.run([
            "docker",
            "build",
            "-t",
            args.docker_image,
            "-f",
            args.docker_file,
        ] + remove_trailing_dashes(args.build_args) + [
            ".",
        ],
                 cwd=args.build_dir)  # noqa
    finally:
        _remove_extra_dirs(args.build_dir, args.extra_dirs)



"""Tests for config.py."""

import pytest
import os

from ..command_path import get_command_dirs_from_config
from ..config_io import ConfigIO


class TestCommandPaths:  # noqa
    @pytest.fixture
    def simple_config(self, tmpdir):
        foo_dir = os.path.join(str(tmpdir), 'foo')
        return {
            'ROOT': {
                'command_path': [
                    os.path.join(foo_dir, '*'),
                ],
                'command_path_exclude': [
                    os.path.join(foo_dir, 'bar'),
                ],
            }
        }

    @pytest.fixture
    def create_config(self, tmpdir, simple_config):
        """Create a config file in the tmpdir."""
        configIO = ConfigIO(str(tmpdir))  # noqa
        configIO.save(simple_config)

    @pytest.fixture
    def create_command_dirs(self, tmpdir):
        """Create a config file in the tmpdir."""
        os.mkdir(os.path.join(str(tmpdir), "foo"))
        os.mkdir(os.path.join(str(tmpdir), "foo", "bar"))
        os.mkdir(os.path.join(str(tmpdir), "foo", "foobar"))

    @pytest.mark.usefixtures("create_config", "create_command_dirs")
    def test_command_path(self, tmpdir):  # noqa
        config = ConfigIO(str(tmpdir)).load()
        command_dirs = get_command_dirs_from_config(config)
        expected_dir = os.path.join(str(tmpdir), "foo", "foobar")
        assert [expected_dir] == command_dirs



import os
from argparse import ArgumentParser

from dodo_commands.dependencies.get import argcomplete, funcy
from dodo_commands.framework.config_layers import get_conflicts_in_layer_paths
from dodo_commands.framework.funcy import (keep_truthy, map_with, remove_if,
                                           str_split_at)

distinct, flatten = funcy.distinct, funcy.flatten


def handle_arg_complete(command_names, inferred_command_names, command_aliases,
                        layer_props_by_layer_name):
    def get_args():
        return os.environ['COMP_LINE'].split()

    def get_prefix_and_command(args):
        return str_split_at(args[1], args[1].rfind('.') + 1)

    input_args = get_args()
    full_command_name = input_args[1]
    # [arg]
    (command_prefix, command_name) = get_prefix_and_command(input_args)

    def get_used_layer_names(command_prefix):
        return keep_truthy()(command_prefix.split('.'))

    used_layer_names = get_used_layer_names(command_prefix)

    def get_commands():
        return distinct(command_names + inferred_command_names +
                        command_aliases)

    def get_choices(commands):
        return [(command_prefix + x) for x in commands]

    commands = get_commands()  # [command_name]
    choices = get_choices(commands)  # [choice]

    def get_possible_layer_names():
        return layer_props_by_layer_name.keys()

    def is_already_used(layer_name):
        return layer_name in used_layer_names

    def is_conflicting(layer_name):
        used_layer_propss = map_with(layer_props_by_layer_name)(
            used_layer_names)
        used_layer_paths = [x.target_path for x in used_layer_propss]

        layer_path = layer_props_by_layer_name[layer_name].target_path
        return get_conflicts_in_layer_paths(used_layer_paths + [layer_path])

    def map_to_choices(layer_name):
        return [command_prefix + layer_name + "." + x for x in commands]

    x = get_possible_layer_names()
    # [layer_name]
    x = remove_if(is_already_used)(x)
    # [layer_name]
    x = remove_if(is_conflicting)(x)
    # [layer_name]
    new_choices = map_with(map_to_choices)(x)
    # [[new_choice]]
    choices += flatten(new_choices)

    if full_command_name not in choices:
        parser = ArgumentParser()
        parser.add_argument('command', choices=choices)
        argcomplete.autocomplete(parser)

    os.environ['COMP_LINE'] = ' '.join(input_args[:1] + input_args[2:])
    os.environ['COMP_POINT'] = str(
        int(os.environ['COMP_POINT']) - (len(full_command_name) + 1))

    return command_name



from importlib import import_module
import fnmatch
import os

from dodo_commands.framework.command_path import (get_command_dirs_from_config,
                                                  extend_sys_path)


def uses_decorator(config, command_name, decorator_name):
    patterns = (config.get('ROOT', {}).get('decorators',
                                           {}).get(decorator_name, []))
    approved = [
        pattern for pattern in patterns if not pattern.startswith("!")
        and fnmatch.filter([command_name], pattern)
    ]
    rejected = [
        pattern for pattern in patterns if pattern.startswith("!")
        and fnmatch.filter([command_name], pattern[1:])
    ]
    return len(approved) and not len(rejected)


def get_decorators(command_name, config):
    result = []
    # should be returned as the last item in the list
    confirm_decorator = None
    for name, directory in _all_decorators(config).items():
        decorator = _load_decorator(name, directory)
        if decorator.is_used(config, command_name, name):
            if name == 'confirm':
                confirm_decorator = decorator
            else:
                result.append(decorator)

    if confirm_decorator:
        result.append(confirm_decorator)

    return result


def _load_decorator(name, directory):
    """Load and return decorator class in module with given name."""
    return import_module(directory + "." + name).Decorator()


def _all_decorators(config):
    """Returns a mapping from decorator name to its directory."""
    command_dirs = get_command_dirs_from_config(config)
    extend_sys_path(command_dirs)
    result = {}
    for item in command_dirs:
        try:
            module_path = os.path.basename(item) + ".decorators"
            module = import_module(module_path)
            for decorator in os.listdir(module.__path__[0]):
                name, ext = os.path.splitext(decorator)
                if ext == '.py' and name != '__init__':
                    result[name] = module_path
        except ImportError:
            continue
    return result



import sys

from dodo_commands.framework import execute_from_command_line
from dodo_commands.framework.global_config import create_global_config


def main():  # noqa
    create_global_config()
    execute_from_command_line(sys.argv)



import os
import sys

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.util import query_yes_no

local = plumbum.local


def _count_confirm_in_argv():
    result = 0
    for arg in sys.argv:
        if arg == '--confirm':
            result += 1
        if arg.startswith('-') and not arg.startswith('--'):
            result += arg.count('c')
        if arg == '--':
            break
    return result


def _ask_confirmation(args, cwd, is_echo, is_confirm):
    """Ask the user whether to continue with executing func."""
    def to_str():
        return args.to_str(slash=args.children)

    if is_echo:
        print(to_str())
        return False

    if is_confirm:
        print("(%s) %s" % (cwd, to_str()))
        if not query_yes_no("confirm?"):
            return False
        else:
            print("")

    return True


def _is_echo(command_line_args):
    return getattr(command_line_args, 'echo', False)


def _is_confirm(command_line_args):
    return (getattr(command_line_args, 'confirm', False)
            or _count_confirm_in_argv()
            or os.environ.get('__DODO_UNIVERSAL_CONFIRM__', None) == '1')


def _is_echo(command_line_args):
    return getattr(command_line_args, 'echo', False)


class Decorator:  # noqa
    def is_used(self, config, command_name, decorator_name):
        return True

    def add_arguments(self, parser):  # noqa
        parser.add_argument(
            '-c',
            '--confirm',
            action='count',
            help="Confirm each performed action before its execution." +
            " Use twice (e.g. dodo foo -cc) to indicate that also nested calls must be confirmed."
        )

        parser.add_argument('-e',
                            '--echo',
                            action='store_true',
                            help="Print all commands instead of running them")

    def modify_args(self, command_line_args, args_tree_root_node, cwd):  # noqa
        if _count_confirm_in_argv() > 1:
            local.env['__DODO_UNIVERSAL_CONFIRM__'] = '1'

        if _is_echo(command_line_args) and not Dodo.safe:
            raise CommandError(
                "The --echo option is not supported for unsafe commands.\n" +
                "Since this command is marked as unsafe, some operations will "
                + "be performed directly, instead of echoed to the console. " +
                "Use --confirm if you wish to execute with visual feedback. ")

        if _is_confirm(command_line_args) and not Dodo.safe:
            if not query_yes_no(
                    "Since this command is marked as unsafe, some operations will "
                    +
                    "be performed without asking for confirmation. Continue?",
                    default="no"):
                sys.exit(1)

        if not _ask_confirmation(args_tree_root_node, cwd or local.cwd,
                                 _is_echo(command_line_args),
                                 _is_confirm(command_line_args)):
            return None, None

        return args_tree_root_node, cwd



"""Tests for all standard commands."""
import os
import shutil
import sys

from dodo_commands.dependencies import (yaml_round_trip_dump,
                                        yaml_round_trip_load)


class TestConfigIO:  # noqa
    @classmethod
    def _load_config(cls, config_filename):
        with open(config_filename) as f:
            return yaml_round_trip_load(f.read())

    @classmethod
    def _write_config(cls, config, config_filename):
        with open(config_filename, 'w') as f:
            f.write(yaml_round_trip_dump(config))

    @classmethod
    def _set_config_version(cls, config_filename, version):
        config = cls._load_config(config_filename)
        config['ROOT']['version'] = version
        cls._write_config(config, config_filename)

    @classmethod
    def _set_required_dodo_version(cls, config_filename, version):
        config = cls._load_config(config_filename)
        config['ROOT']['required_dodo_commands_version'] = version
        cls._write_config(config, config_filename)

    @classmethod
    def _clear_layers(cls, config_filename):
        config = cls._load_config(config_filename)
        config['LAYERS'] = []
        cls._write_config(config, config_filename)

    def test_commands(self):  # noqa
        from dodo_commands.dependencies.get import plumbum
        local = plumbum.local

        dodo_test_dir = os.path.expanduser('~/projects/dodo_test')
        dodo_test_env_dir = os.path.expanduser(
            '~/.dodo_commands/envs/dodo_test')
        skip_install = False

        if not skip_install:
            if os.path.exists(dodo_test_dir):
                shutil.rmtree(dodo_test_dir)
            if os.path.exists(dodo_test_env_dir):
                shutil.rmtree(dodo_test_env_dir)

            from dodo_commands.dodo import main
            sys.argv = [
                'dodo', 'env', '--create', '--create-python-env', 'dodo_test'
            ]
            main()

        dodo = local[os.path.join(dodo_test_dir, '.env/bin/dodo')]
        if not skip_install:
            dodo('bootstrap', 'src', 'extra/dodo_commands/res', '--force',
                 '--git-url',
                 'https://github.com/mnieber/dodo_commands_tutorial.git')

        config_dir = os.path.join(dodo_test_dir, '.dodo_commands')
        config_filename = os.path.join(config_dir, 'config.yaml')

        # dodo which
        assert "dodo_test" == dodo('which')[:-1]
        assert dodo_test_dir == dodo('which', 'project')[:-1]
        assert os.path.join(dodo_test_dir, 'src') == dodo('which', 'src')[:-1]
        assert config_dir == dodo('which', 'res')[:-1]
        assert config_filename == dodo('which', '--config')[:-1]
        assert 'confirm, debugger, docker, pause' == dodo(
            'which', '--decorators')[:-1]

        # dodo cd
        assert "cd " + dodo_test_dir == dodo('cd')[:-1]

        # dodo check-version --config
        self._set_config_version(config_filename, '0.1.0')
        assert "Configuration needs update (0.1.0 < 1.0.0)" in dodo(
            'check-version', '--config')[:-1]
        self._set_config_version(config_filename, '1.0.0')

        # dodo check-version --dodo
        self._set_required_dodo_version(config_filename, '10000.0.0')
        assert "The dodo_commands package needs to be upgraded" in dodo(
            'check-version', '--dodo')[:-1]
        self._set_required_dodo_version(config_filename,
                                        dodo('--version')[:-1])
        assert "" == dodo('check-version', '--dodo')[:-1]

        # dodo diff
        result = dodo('diff', '.', '--echo').replace('\n', '')
        assert "diff %s/src/extra/dodo_commands/res %s/." % (
            dodo_test_dir, config_dir) == result

        # dodo docker
        dodo('layer', 'docker', 'on')
        result = dodo('docker', '*', '--name', 'foo',
                      '--echo').replace('\n',
                                        '').replace('  \\',
                                                    '').replace('  ', ' ')
        assert "docker run --name=foo --interactive --tty --rm --volume=%s/log:/var/log --workdir=/ dodo_tutorial:1604 /bin/bash" % dodo_test_dir == result

        # dodo docker-build
        result = dodo('docker-build', '--echo',
                      'base').replace('\n', '').replace('  \\', '')
        assert "docker build -t dodo_tutorial:1604 -f Dockerfile ." == result

        # dodo autostart
        autostart_file = os.path.expanduser('~/.dodo_commands_autostart')
        autostart_bak = autostart_file + '.bak'
        if os.path.exists(autostart_file):
            os.rename(autostart_file, autostart_bak)
        dodo('autostart', 'on')
        assert os.path.exists(autostart_file)
        dodo('autostart', 'off')
        assert not os.path.exists(autostart_file)
        if os.path.exists(autostart_bak):
            os.rename(autostart_bak, autostart_file)

        # dodo layer
        self._clear_layers(config_filename)
        dodo('layer', 'debug', 'on')
        assert self._load_config(config_filename)['LAYERS'] == [
            'debug.on.yaml'
        ]
        dodo('layer', 'debug', 'off')
        assert self._load_config(config_filename)['LAYERS'] == [
            'debug.off.yaml'
        ]

        # dodo print-config
        assert "CMAKE_INSTALL_PREFIX: %s/install" % dodo_test_dir in dodo(
            "print-config")

        # dodo drop-in
        tutorial_commands_dir = os.path.join(
            dodo_test_dir, "src/extra/dodo_commands/dodo_tutorial_commands")
        drop_in_dir = os.path.join(tutorial_commands_dir, "drop-in")
        os.mkdir(drop_in_dir)
        with open(os.path.join(drop_in_dir, "bar.txt"), "w") as ofs:
            ofs.write("bar")
        target_bar_path = os.path.join(config_dir, 'drops',
                                       'dodo_tutorial_commands', 'bar.txt')
        assert not os.path.exists(target_bar_path)
        dodo('drop-in', 'dodo_tutorial_commands')
        assert os.path.exists(target_bar_path)



from dodo_commands.framework.command_error import CommandError  # noqa
from dodo_commands.framework.config_arg import ConfigArg  # noqa
from dodo_commands.framework.decorator_scope import DecoratorScope  # noqa
from dodo_commands.framework.singleton import Dodo  # noqa
from dodo_commands.framework.util import remove_trailing_dashes  # noqa



from argparse import ArgumentParser

from dodo_commands import DecoratorScope, Dodo

from dodo_docker_commands.decorators.docker import Decorator as DockerDecorator


def _choices():
    choices = []
    for key in Dodo.get_config('/DOCKER_OPTIONS', {}).keys():
        keys = [key] if isinstance(key, str) else key
        for x in keys:
            if x not in choices and not x.startswith('!'):
                choices.append(str(x))
    return choices


def _args():
    parser = ArgumentParser(
        description='Opens a shell in the docker container.')
    parser.add_argument(
        'service',
        choices=_choices(),
        help=("Use this key to look up the docker options in /DOCKER_OPTIONS"))
    parser.add_argument(
        '--image',
        choices=Dodo.get_config('/DOCKER_IMAGES', {}).keys(),
        help=("Use the docker image stored under this key in /DOCKER_IMAGES"))
    parser.add_argument('--image-name',
                        help=("Use the docker image with this name"))
    parser.add_argument(
        '--name', help=("Override the name of the started docker container"))
    parser.add_argument('--command')
    args = Dodo.parse_args(parser)
    return args


if Dodo.is_main(__name__):
    args = _args()

    docker_options = DockerDecorator.merged_options(Dodo.get_config,
                                                    args.service)

    if args.image:
        docker_options['image'] = Dodo.get_config(
            '/DOCKER_IMAGES/%s/image' % args.image, args.image)
    elif args.image_name:
        docker_options['image'] = args.image_name

    if args.name:
        docker_options['name'] = args.name
    else:
        docker_options['name'] = args.service

    Dodo.get_config('/DOCKER')['options'] = {Dodo.command_name: docker_options}

    with DecoratorScope('docker'):
        Dodo.run([args.command] if args.command else [],
                 cwd=docker_options.get("cwd", "/"))



"""Decorates command lines with docker arguments."""
import os
from fnmatch import fnmatch

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.config import merge_into_config
from dodo_commands.framework.decorator_utils import uses_decorator

local = plumbum.local


def _is_tuple(x):
    return hasattr(x, "__len__") and not isinstance(x, type(str()))


# Resp: reads docker related settings from the docker_config and uses
# them to modify the current args_tree_root_node.


class Decorator:  # noqa
    def is_used(self, config, command_name, decorator_name):
        return uses_decorator(config, command_name, decorator_name)

    @classmethod
    def _add_docker_volume_list(cls, docker_config, root_node):
        volume_list = docker_config.get('volume_list', [])
        volume_map = docker_config.get('volume_map', {})

        volume_map_strict = docker_config.get('volume_map_strict', {})
        for key, val in volume_map_strict.items():
            if not os.path.exists(key):
                raise CommandError("Path in volume_map_strict not found: %s" %
                                   key)
            volume_map[key] = val

        options = (
            ['--volume=%s:%s' % (x, x) for x in volume_list] +
            ['--volume=%s:%s' % key_val for key_val in volume_map.items()])
        for x in options:
            root_node['volume'].append(x)

    @classmethod
    def _add_docker_publish_list(cls, docker_config, root_node):
        publish_list = docker_config.get('publish_list', [])
        publish_map = docker_config.get('publish_map', {})
        options = (
            ['--publish=%s:%s' % (x, x) for x in publish_list] +
            ['--publish=%s:%s' % key_val for key_val in publish_map.items()])
        for x in options:
            root_node['publish'].append(x)

    @classmethod
    def _add_docker_volumes_from_list(cls, docker_config, root_node):
        volumes_from_list = docker_config.get('volumes_from_list', [])
        options = (['--volumes-from=%s' % x for x in volumes_from_list])
        for x in options:
            root_node['volumes-from'].append(x)

    @classmethod
    def _add_docker_variable_list(cls, docker_config, root_node):
        variable_list = docker_config.get('variable_list', [])
        variable_map = docker_config.get('variable_map', {})
        options = (
            ['--env=%s' % x for x in variable_list] +
            ['--env=%s=%s' % key_val for key_val in variable_map.items()])
        for x in options:
            root_node['env'].append(x)

    @classmethod
    def _add_linked_container_list(cls, docker_config, root_node):
        options = ['--link=%s' % x for x in docker_config.get('link_list', [])]
        for x in options:
            root_node['link'].append(x)

    @classmethod
    def _add_interactive(cls, docker_config, docker_node):
        if docker_config.get('is_interactive', True):
            docker_node['basic'].append('--interactive')
            docker_node['basic'].append('--tty')

    @classmethod
    def _add_workdir(cls, docker_config, docker_node, cwd):
        if cwd and not ('workdir' in docker_config
                        and not docker_config['workdir']):
            docker_node['workdir'].append('--workdir=' + cwd)
        elif docker_config.get('workdir', ''):
            docker_node['workdir'].append('--workdir=' +
                                          docker_config.get('workdir'))

    @classmethod
    def _add_user(cls, docker_config, docker_node):
        if docker_config.get('user', False):
            docker_node['basic'].extend(['--user', docker_config['user']])

    @classmethod
    def _add_environment_vars(cls, get_config, docker_config, docker_node):
        for key_val in get_config('/ENVIRONMENT/variable_map', {}).items():
            docker_node['env'].append("--env=%s=%s" % key_val)

    @classmethod
    def _add_extra_options(cls, docker_config, docker_node):
        for x in docker_config.get('extra_options', []):
            docker_node['other'].append(x)

    @classmethod
    def merged_options(cls, get_config, command_name):
        merged = {}
        docker_options = get_config('/DOCKER_OPTIONS', {})
        for patterns, docker_config in docker_options.items():
            should_merge = False
            for pattern in (patterns if _is_tuple(patterns) else [patterns]):
                if pattern.startswith('!'):
                    should_merge = should_merge and not fnmatch(
                        command_name, pattern[1:])
                else:
                    should_merge = should_merge or fnmatch(
                        command_name, pattern)
            if should_merge:
                merge_into_config(merged, docker_config)
        return merged

    @classmethod
    def docker_node(cls, get_config, command_name, cwd):
        merged = cls.merged_options(get_config, command_name)

        image, container = merged.get('image'), merged.get('container')

        if image:
            docker_node = ArgsTreeNode("docker", args=['docker', 'run'])
            docker_node.add_child(ArgsTreeNode("name"))
            docker_node.add_child(ArgsTreeNode("basic"))
            docker_node.add_child(ArgsTreeNode("link"))
            docker_node.add_child(ArgsTreeNode("publish"))
            docker_node.add_child(ArgsTreeNode("volume", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("volumes-from"))
            docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("other"))
            docker_node.add_child(ArgsTreeNode("workdir"))
            docker_node.add_child(ArgsTreeNode("image"))

            cls._add_linked_container_list(merged, docker_node)
            cls._add_docker_publish_list(merged, docker_node)
            cls._add_docker_volume_list(merged, docker_node)
            cls._add_docker_volumes_from_list(merged, docker_node)
            cls._add_docker_variable_list(merged, docker_node)
            cls._add_workdir(merged, docker_node, cwd)
            cls._add_interactive(merged, docker_node)
            cls._add_user(merged, docker_node)
            cls._add_environment_vars(get_config, merged, docker_node)
            cls._add_extra_options(merged, docker_node)

            if merged.get('rm', True):
                docker_node['basic'].append('--rm')

            if merged.get('is_daemon', False):
                docker_node['basic'].append('-d')

            name = merged.get('name', command_name)
            if name:
                docker_node["name"].append("--name=%s" % name)

            docker_node['image'].append(image)
        elif container:
            docker_node = ArgsTreeNode("docker", args=['docker', 'exec'])
            docker_node.add_child(ArgsTreeNode("basic"))
            docker_node.add_child(ArgsTreeNode("env", is_horizontal=False))
            docker_node.add_child(ArgsTreeNode("other"))
            docker_node.add_child(ArgsTreeNode("workdir"))
            docker_node.add_child(ArgsTreeNode("container"))

            cls._add_interactive(merged, docker_node)
            cls._add_docker_variable_list(merged, docker_node)
            cls._add_workdir(merged, docker_node, cwd)
            cls._add_user(merged, docker_node)
            cls._add_environment_vars(get_config, merged, docker_node)
            cls._add_extra_options(merged, docker_node)
            docker_node['container'].append(container)
        else:
            raise CommandError(
                "No docker image or container found in /DOCKER_OPTIONS for command %s"
                % command_name)

        return docker_node, image

    def add_arguments(self, parser):  # noqa
        pass

    def modify_args(self, command_line_args, args_tree_root_node,
                    cwd):  # override
        docker_node, _ = self.docker_node(Dodo.get_config, Dodo.command_name,
                                          cwd)

        docker_node.add_child(args_tree_root_node)
        return docker_node, local.cwd



import os
import sys
from argparse import ArgumentParser

from dodo_commands import CommandError, Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.choice_picker import ChoicePicker
from dodo_commands.framework.util import exe_exists

tmux = plumbum.cmd.tmux


def _session_id(category):
    return os.path.expandvars('$USER') + "_" + category


def _normalize(category):
    return category.replace(' ', '-')


def _args():
    command_map = Dodo.get_config('/MENU/commands', {})

    parser = ArgumentParser()
    parser.add_argument('category',
                        choices=['all'] +
                        list([_normalize(x) for x in command_map.keys()]),
                        nargs='?')
    parser.add_argument('--tmux', action='store_true')
    parser.add_argument('--list', action='store_true')
    parser.add_argument('--run', type=int, nargs='?', const=-1)

    args = Dodo.parse_args(parser)
    args.category = args.category or 'all'
    args.run = -1 if args.run is None else args.run
    args.command_map = command_map

    return args


def _create_tmux_window(session_id):
    # Create tmux session
    tmux('-2', 'new-session', '-d', '-s', session_id)

    # Create a tmux window
    tmux('new-window', '-t', '%s:1' % session_id, '-n', 'Logs')


def _get_categories(command_map, category):
    return [
        x for x in command_map
        if category == 'all' or _normalize(x) == category
    ]


def _get_commands_and_labels(command_map, category):
    categories = _get_categories(command_map, category)
    label_size = 0
    for category in categories:
        label_size = max(label_size, len(category))
    label_prefix = "%0" + str(label_size) + "s"

    commands, labels = [], []
    for category in categories:
        for command in command_map[category]:
            commands.append(command)
            format_string = "%02s [" + label_prefix + "] - %s"
            labels.append(format_string %
                          (str(len(commands)), category, command))

    return commands, labels


def _get_selected_commands(commands, labels, allow_free_text=False):
    class Picker(ChoicePicker):
        def print_choices(self, choices):
            print()
            for idx, label in enumerate(choices):
                print(label)
            print()

        def question(self):
            return (
                'Select one or more commands (e.g. 1,3-4)%s or type 0 to exit: '
                % (', or type a command,' if allow_free_text else ''))

        def on_invalid_index(self, index):
            if index == 0:
                sys.exit(0)

    picker = Picker(commands, allow_free_text=allow_free_text, labels=labels)
    picker.pick()
    return [picker.free_text] if picker.free_text else picker.get_choices()


if Dodo.is_main(__name__):
    args = _args()
    check_exists = Dodo.get_config('/MENU/check_exists', '/')
    if not os.path.exists(check_exists):
        raise CommandError("Path %s does not exist" % check_exists)

    commands, labels = _get_commands_and_labels(args.command_map,
                                                args.category)

    if not commands:
        raise CommandError(
            "No commands were found in the /MENU configuration key")

    if args.list:
        print()
        for label in labels:
            print(label)
    elif args.tmux:
        if not exe_exists('tmux'):
            raise CommandError('Tmux is not installed on this sytem.')

        session_id = _session_id(args.category)
        has_session = False
        try:
            sessions = tmux('ls')
            for session in sessions.split('\n'):
                has_session = has_session or session.startswith(
                    '%s:' % session_id)
        except:  # noqa
            pass

        if not has_session:
            _create_tmux_window(session_id)
            tmux('send-keys', 'dodo menu --tmux %s' % args.category, 'C-m')
            # Attach to tmux session
            # HACK: why does this only work via Dodo.run?
            Dodo.run(['tmux', '-2', 'attach-session', '-t', session_id], )
        else:
            while True:
                selected_commands = _get_selected_commands(
                    commands, labels, allow_free_text=True)
                for command in selected_commands:
                    print(command)
                    tmux('split-window', '-v')
                    tmux('send-keys', command, 'C-m')

                # Set default window
                tmux('select-pane', '-t', '0')
                tmux('select-layout', 'tile')
    else:
        selected_commands = (_get_selected_commands(commands, labels)
                             if args.run == -1 else [commands[args.run - 1]])
        for command in selected_commands:
            Dodo.run(['bash', '-c', command])



import sys
from argparse import ArgumentParser

from dodo_commands import Dodo
from dodo_commands.dependencies.get import plumbum
from dodo_commands.framework.choice_picker import ChoicePicker

docker = plumbum.cmd.docker


def _args():
    parser = ArgumentParser()
    parser.add_argument('--find')
    parser.add_argument('--ip', action='store_true')
    args = Dodo.parse_args(parser)
    args.names = []
    return args


def _containers():
    return [x for x in docker("ps", "--format", "{{.Names}}").split('\n') if x]


if Dodo.is_main(__name__):
    args = _args()

    if args.find:
        for container in _containers():
            if args.find in container:
                args.names = [container]
                break
    else:

        class Picker(ChoicePicker):
            def print_choices(self, choices):
                print("0 - exit")
                for idx, container in enumerate(choices):
                    print("%d - %s" % (idx + 1, container))
                print()

            def question(self):
                return "Select a container: "

            def on_invalid_index(self, index):
                if index == 0:
                    sys.exit(0)

        picker = Picker(_containers())
        picker.pick()
        args.names = picker.get_choices()

    ip_args = [
        "-f", "{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}"
    ] if args.ip else []
    for container in args.names:
        Dodo.run(['docker', 'inspect', *ip_args, container], )



import argparse
import os
import sys

from dodo_commands.dependencies.get import argcomplete, plumbum
from dodo_commands.framework.args_tree import ArgsTreeNode
from dodo_commands.framework.command_error import CommandError
from dodo_commands.framework.config_arg import add_config_args
from dodo_commands.framework.config_expander import Key, KeyNotFound
from dodo_commands.framework.container.container import Container
from dodo_commands.framework.decorator_utils import get_decorators
from dodo_commands.framework.util import classproperty


class Dodo:
    safe = True
    parser = argparse.ArgumentParser()

    _args = None
    _container = None

    @classmethod
    def get_container(cls):
        if cls._container is None:
            cls._container = Container()
            cls._container.run_actions()

        return cls._container

    @classproperty
    def command_name(cls):  # noqa
        return cls.get_container().command_line.command_name

    @classproperty
    def args(cls):  # noqa
        if cls._args is None:
            cls.parse_args(cls.parser)
        return cls._args

    @classproperty
    def command_name(cls):  # noqa
        return cls.get_container().command_line.command_name

    @classmethod
    def get(cls, key='', default_value="__not_set_234234__"):  # noqa
        try:
            return Key(cls.get_container().config.config, key).get()
        except KeyNotFound:
            if default_value == "__not_set_234234__":
                raise
            return default_value

    @classmethod
    def is_main(cls, name, safe=None):
        if safe is not None:
            cls.safe = safe

        if name == '__main__':
            return True

        commands = cls.get_container().commands
        command_map_item = commands.command_map.get(cls.command_name, None)
        if not command_map_item:
            return False

        script_path = command_map_item.package_path + '.' + cls.command_name
        return name == script_path

    @classmethod
    def parse_args(cls, parser, config_args=None):
        parser.add_argument('--traceback',
                            action='store_true',
                            help=argparse.SUPPRESS)

        add_config_args(parser, cls.get_config(), config_args)

        for decorator in get_decorators(cls.command_name, cls.get_config()):
            decorator.add_arguments(parser)

        if cls.get_container().command_line.is_help:
            parser.print_help()
            sys.exit(0)

        argcomplete.autocomplete(parser)

        args = cls.get_container().command_line.input_args
        parser.prog = os.path.basename(args[0])
        if len(args) > 1:
            parser.prog += " %s" % args[1]
        cls._args = parser.parse_args(args[2:])

        if config_args:
            for config_arg in config_args:
                key = Key(cls.get_config(), config_arg.config_key)
                if key.exists():
                    setattr(cls._args, config_arg.arg_name, key.get())

        return cls._args

    @classmethod
    def run(cls, args, cwd=None, quiet=False, capture=False):
        args_tree_root_node = ArgsTreeNode('original_args', args=args)
        for decorator in get_decorators(cls.command_name, cls.get_config()):
            args_tree_root_node, cwd = decorator.modify_args(
                cls._args, args_tree_root_node, cwd)
            if args_tree_root_node is None:
                return False

        if cwd and not os.path.exists(cwd):
            raise CommandError("Directory not found: %s" % cwd)

        with plumbum.local.cwd(cwd or plumbum.local.cwd):
            flat_args = args_tree_root_node.flatten()
            func = plumbum.local[flat_args[0]][flat_args[1:]]
            variable_map = cls.get_config('/ENVIRONMENT/variable_map', {})
            with plumbum.local.env(**variable_map):
                if capture:
                    return func()
                try:
                    func & plumbum.FG
                    return True
                except plumbum.ProcessExecutionError:
                    if not quiet:
                        print(
                            "\nDodo Commands error while running this command:"
                        )
                        print("\n\n%s" % func)
                    return False

    # add alias for the legacy 'runcmd' method
    runcmd = run
    get_config = get
